/*
 * Copyright (C) 2005 Andre Noll <maan@systemlinux.org>
 *
 *     This program is free software; you can redistribute it and/or modify
 *     it under the terms of the GNU General Public License as published by
 *     the Free Software Foundation; either version 2 of the License, or
 *     (at your option) any later version.
 *
 *     This program is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *     GNU General Public License for more details.
 *
 *     You should have received a copy of the GNU General Public License
 *     along with this program; if not, write to the Free Software
 *     Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111, USA.
 */

/*
 * Based in parts on aplay.c from the alsa-utils-1.0.8 package,
 * Copyright (c) by Jaroslav Kysela <perex@suse.cz>, which is
 * based on the vplay program by Michael Beck.
 */

#define WAV_HEADER_LEN 44
#include "para.h"
#include "play.cmdline.h"
#include <alsa/asoundlib.h>

enum {	E_BROKEN_CONF,		/* Broken configuration for this PCM */
	E_ACCESS_TYPE,		/* Access type not available */
	E_SAMPLE_FORMAT,	/* Sample format not available */
	E_CHANNEL_COUNT,	/* Channels count not available */
	E_HW_PARAMS,		/* Unable to install hw params */
	E_SW_PARAMS,		/* Unable to install sw params */
	E_BAD_PERIOD,		/* Can't use period equal to buffer size */
	E_GET_XFER,		/* Unable to obtain xfer align */
	E_SET_XFER,		/* snd_pcm_sw_params_set_xfer_align failed */
	E_MEM,			/* not enough memory */
	E_READ,			/* read error */
	E_WRITE,		/* write error */
	E_PIPE,			/* write to pipe with other side closed */
	E_PCM_OPEN,		/* unable to open pcm */
	E_INFO,			/* pcm info error */
	E_GET_BUFFER_TIME,	/* snd_pcm_hw_params_get_buffer_time_max failed */
	E_SET_BUFFER_TIME,	/* snd_pcm_hw_params_set_buffer_time_near failed */
	E_SET_RATE,		/* snd_pcm_hw_params_set_rate_near failed */
	E_START_THRESHOLD,	/* snd_pcm_sw_params_set_start_threshold failed */
	E_STOP_THRESHOLD,	/* snd_pcm_sw_params_set_stop_threshold failed */
	E_LOG,			/* snd_output_stdio_attach failed */
	E_SYNTAX		/* could not parse start_time option */
};

#define FORMAT SND_PCM_FORMAT_S16_LE

static snd_pcm_t *handle;
static unsigned char *audiobuf;
static snd_pcm_uframes_t chunk_size;
static size_t bytes_per_frame;
static struct timeval *start_time;
static struct gengetopt_args_info conf;

static ssize_t read_chunk(unsigned char *buf, size_t count)
{
	ssize_t result = 0, res;

	while (count > 0) {
		if ((res = read(STDIN_FILENO, buf, count)) == 0)
			break;
		if (res < 0)
			return result > 0 ? result : res;
		count -= res;
		result += res;
		buf = buf + res;
	}
	return result;
}

static void set_params(void)
{
	snd_pcm_hw_params_t *hwparams;
	snd_pcm_sw_params_t *swparams;
	snd_pcm_uframes_t buffer_size, xfer_align, start_threshold,
		stop_threshold;
	unsigned buffer_time = 0;
	int err;

	snd_pcm_hw_params_alloca(&hwparams);
	snd_pcm_sw_params_alloca(&swparams);
	if (snd_pcm_hw_params_any(handle, hwparams) < 0)
		exit(E_BROKEN_CONF);
	if (snd_pcm_hw_params_set_access(handle, hwparams,
			SND_PCM_ACCESS_RW_INTERLEAVED) < 0)
		exit(E_ACCESS_TYPE);
	if (snd_pcm_hw_params_set_format(handle, hwparams, FORMAT) < 0)
		exit(E_SAMPLE_FORMAT);
	if (snd_pcm_hw_params_set_channels(handle, hwparams, conf.channels_arg) < 0)
		exit(E_CHANNEL_COUNT);
	if (snd_pcm_hw_params_set_rate_near(handle, hwparams, &conf.sample_rate_arg, 0) < 0)
		exit(E_SET_RATE);
	err = snd_pcm_hw_params_get_buffer_time_max(hwparams, &buffer_time, 0);
	if (err < 0 || !buffer_time)
		exit(E_GET_BUFFER_TIME);
	if (buffer_time > 500000)
		buffer_time = 500000;
	if (snd_pcm_hw_params_set_buffer_time_near(handle, hwparams,
			&buffer_time, 0) < 0)
		exit(E_SET_BUFFER_TIME);
	if (snd_pcm_hw_params(handle, hwparams) < 0)
		exit(E_HW_PARAMS);
	snd_pcm_hw_params_get_period_size(hwparams, &chunk_size, 0);
	snd_pcm_hw_params_get_buffer_size(hwparams, &buffer_size);
	if (chunk_size == buffer_size)
		exit(E_BAD_PERIOD);
	snd_pcm_sw_params_current(handle, swparams);
	err = snd_pcm_sw_params_get_xfer_align(swparams, &xfer_align);
	if (err < 0 || !xfer_align)
		exit(E_GET_XFER);
	snd_pcm_sw_params_set_sleep_min(handle, swparams, 0);
	snd_pcm_sw_params_set_avail_min(handle, swparams, chunk_size);
	/* round to closest transfer boundary */
	start_threshold = (buffer_size / xfer_align) * xfer_align;
	if (start_threshold < 1)
		start_threshold = 1;
	if (snd_pcm_sw_params_set_start_threshold(handle, swparams,
			start_threshold) < 0)
		exit(E_START_THRESHOLD);
	stop_threshold = buffer_size;
	if (snd_pcm_sw_params_set_stop_threshold(handle, swparams,
			stop_threshold) < 0)
		exit(E_STOP_THRESHOLD);
	if (snd_pcm_sw_params_set_xfer_align(handle, swparams, xfer_align) < 0)
		exit(E_SET_XFER);
	if (snd_pcm_sw_params(handle, swparams) < 0)
		exit(E_SW_PARAMS);
	bytes_per_frame = snd_pcm_format_physical_width(FORMAT) * conf.channels_arg / 8;
}

static ssize_t pcm_write(u_char *data, size_t count)
{
	ssize_t r, result = 0;

	if (count < chunk_size) {
		snd_pcm_format_set_silence(FORMAT, data
			+ count * bytes_per_frame,
			(chunk_size - count) * conf.channels_arg);
		count = chunk_size;
	}
	while (count > 0) {
		r = snd_pcm_writei(handle, data, count);
		if (r == -EAGAIN || (r >= 0 && r < count))
			snd_pcm_wait(handle, 1);
		else if (r == -EPIPE)
			snd_pcm_prepare(handle);
		else if (r < 0)
			exit(E_WRITE);
		if (r > 0) {
			result += r;
			count -= r;
			data += r * bytes_per_frame;
		}
	}
	return result;
}

static void do_initial_delay(void)
{
	struct timeval now, diff;

	if (!conf.start_time_given)
		return;
	gettimeofday(&now, NULL);
	if (tv_diff(start_time, &now, &diff) > 0)
		select(1, NULL, NULL, NULL, &diff);
}

/* play raw data */
static void playback(int loaded)
{
	int r;
	unsigned long written = 0;
	size_t chunk_bytes, l;

	set_params();
	chunk_bytes = chunk_size * bytes_per_frame;
	audiobuf = realloc(audiobuf, chunk_bytes);
	if (!audiobuf)
		exit(E_MEM);
	while (loaded > chunk_bytes) {
		if (pcm_write(audiobuf + written, chunk_size) <= 0)
			return;
		written += chunk_bytes;
		loaded -= chunk_bytes;
	}
	if (written > 0 && loaded > 0)
		memmove(audiobuf, audiobuf + written, loaded);
	l = loaded;
	for (;;) {
		do {
			if (chunk_bytes == l)
				break;
			r = read_chunk(audiobuf + l, chunk_bytes - l);
			if (r < 0)
				exit(E_READ);
			if (r == 0)
				break;
			l += r;
		} while (l < chunk_bytes);
		if (!written && start_time)
			do_initial_delay();
		l = l / bytes_per_frame;
		r = pcm_write(audiobuf, l);
		if (r != l)
			break;
		r = r * bytes_per_frame;
		written += r;
		l = 0;
	}
	snd_pcm_drain(handle);
}

static int check_wave(void)
{
	unsigned char *a = audiobuf;
	if (a[0] != 'R' || a[1] != 'I' || a[2] != 'F' || a[3] != 'F')
		return WAV_HEADER_LEN;
	conf.channels_arg = (unsigned) a[22];
	conf.sample_rate_arg = a[24] + (a[25] << 8) + (a[26] << 16) + (a[27] << 24);
//	fprintf(stderr, "channels: %d, rate: %d\n", conf.channels_arg,
//		conf.sample_rate_arg);
	return 0;
}
int main(int argc, char *argv[])
{
	snd_pcm_info_t *info;
	snd_pcm_stream_t stream = SND_PCM_STREAM_PLAYBACK;
	snd_output_t *log;
	struct timeval tv;
	int err;

	cmdline_parser(argc, argv, &conf);
	if (conf.start_time_given) {
		if (sscanf(conf.start_time_arg, "%lu:%lu",
				&tv.tv_sec, &tv.tv_usec) != 2)
			exit(E_SYNTAX);
		start_time = &tv;
	}
//	fprintf(stderr, "argc=%d, argv[1]=%s\n",argc,  argv[1]);
	snd_pcm_info_alloca(&info);
	if (snd_output_stdio_attach(&log, stderr, 0) < 0)
		exit(E_LOG);
	err = snd_pcm_open(&handle, "plug:swmix", stream, 0);
	if (err < 0)
		exit(E_PCM_OPEN);
	if ((err = snd_pcm_info(handle, info)) < 0)
		exit(E_INFO);
	audiobuf = malloc(WAV_HEADER_LEN);
	err = read_chunk(audiobuf, WAV_HEADER_LEN);
	if (err < 0)
		exit(E_READ);
	playback(check_wave());
	snd_pcm_close(handle);
	free(audiobuf);
	snd_output_close(log);
	snd_config_update_free_global();
	return EXIT_SUCCESS;
}
